// Copyright (C) Julien Lerique 2010 (ljpupfr _at_ gmail _dot_ com)
// 2010 GPL license version 3 (http://www.gnu.org/licenses/gpl.html).
// compilation of xmltool:
// gcc -Os `pkg-config --cflags libxml-2.0 --libs libxml-2.0` \
//  -I/usr/include xmltool.c -o ../xmltool
// and optionaly:
// strip --strip-all --remove-section=.comment --remove-section=.note ../xmltool
 /** 
 * A basic tool to read or modifiy special nodes/attributes of an XML file
 * using libxml with xpath expressions.
 */
#include <stdio.h>
#include <string.h>
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>

typedef enum 
{
  COUNT,
  GET,
  PATH,
  REMOVE,  
  SET,
  COPY_AS_LASTCHILD,
  COPY_AS_NEXTSIBLING,
  COPY_AS_PREVSIBLING,
  NOTHING
} xmltoolAction;

/**
 * usage:
 * @name:		the program name.
 *
 * Prints usage information.
 */
void usage(const char *name) {
  printf( "Usage: %s function [arguments]\n"
          "\n"
          "%s is small command line tool to retrieve and/or modify\n"
          "informations of an xml file.\n"
          "\n"
          "Defined functions are:\n"
          "   count : count the number of nodes matching a xpath expression\n"
          "   get : return the result of a given xpath expression\n"
          "   path : return full xpath(s) computed from a xpath expression\n"
          "   remove : remove node matching a xpath expression from file\n" 
          "   set : change the value of given nodes to a new value\n"
          "   copy_as_lastchild : copy a node as the last child of the\n"
          "       destination node\n"
          "   copy_as_nextsibling : copy a node as the next sibling of the\n"
          "       destination node\n" 
          "   copy_as_prevsibling : : copy a node as the previous sibling of\n"
          "       the destination node\n"
          "\n"
          "Arguments for functions count/get/path/remove are:\n"
          "   <xml-file> <xpath-expr>\n"
          "Arguments for function set are:\n"
          "   <xml-file> <xpath-expr> <xpath-new-value>\n"
          "\n"
          "Arguments for functions starting with copy_as_ are:\n"
          "   <xml-file-source> <xpath-expr-source> \n"
          "   <xml-file-dest> <xpath-expr-dest> <extended>\n"
          "As in libxml's xmlCopyNode, set the extended parameter to\n"
          "   1 : for a recursive copy of properties, namespaces, children\n"
          "   2 : to copy only properties and namespaces", 
          name, name);
}

void error_msg(int type, const char* text) {
  printf(
    (type==0)?"Error: unable to parse file \"%s\"\n":(
    (type==1)?"Error: unable to create XPath context of \"%s\"\n":
    "Error: unable to evaluate expression \"%s\"\n"), text);
}

/**
 * execute_expression:
 * @taskType:		the type of action to be performed
 * @filename:		the input XML filename.
 * @xpathExpr:		the xpath expression to work on.
 * @xpathValue:	the new node value, if required
 *
 * Parses input XML file, evaluates XPath expression and perform the requested
 * action.
 * 
 * Returns 0 on success and a negative value otherwise.
 */
int execute_expression(xmltoolAction type, const char* filename,
      const xmlChar* xpathExpr, const xmlChar* xpathValue) {
    
  xmlDocPtr doc;
  xmlXPathContextPtr xpathCtx; 
  xmlXPathObjectPtr xpathObj; 
  
  xmlNodePtr cur=NULL;
  int i=0;
  int nodeNb=0;
  int isAttribute=0;
  int res=-1;
  
  /* Load XML document */
  doc = xmlParseFile(filename);
  if (!doc) {
    error_msg( 0, filename);
  } else {
    /* Create xpath evaluation context */
    xpathCtx = xmlXPathNewContext(doc);
    if (!xpathCtx) {
      error_msg( 1, filename);
    } else {
      /* Evaluate xpath expression */
      xpathObj = xmlXPathEvalExpression(xpathExpr, xpathCtx);
      if (!xpathObj) {
        error_msg( 2, xpathExpr);
      } else {
        
        /* Results */
        nodeNb=(xpathObj->nodesetval) ? xpathObj->nodesetval->nodeNr : 0;
        isAttribute=xmlStrchr(xmlGetNodePath(cur),'@') ? 1 : 0;
        res=0;
        
        switch(type){
          case COUNT:
            printf("%d\n",nodeNb);
            break;
          case GET:
            for(i = 0; i < nodeNb; ++i) {
              cur = xpathObj->nodesetval->nodeTab[i];
              if(isAttribute)
                printf("%s\n",xmlGetProp(cur->parent,cur->name));
              else
                printf("%s\n",xmlNodeGetContent(cur));
            }
            break;
          case PATH:
            for(i = 0; i < nodeNb; ++i) 
              printf("%s\n",xmlGetNodePath(xpathObj->nodesetval->nodeTab[i]));
            break;
          case REMOVE:
            for(i = 0; i < nodeNb; ++i) {
              cur = xpathObj->nodesetval->nodeTab[i];
              if(isAttribute)
                xmlUnsetProp(cur->parent,cur->name);
              else
                xmlUnlinkNode(cur);
            }
            xmlSaveFormatFile(filename, doc, 1);
            break;
          case SET:
            for(i = 0; i < nodeNb; ++i) {
              cur = xpathObj->nodesetval->nodeTab[i];
              if(isAttribute)
                xmlSetProp(cur->parent,cur->name, xpathValue);
              else
                xmlNodeSetContent(cur, xpathValue);
            }
            xmlSaveFormatFile(filename, doc, 1);
            break;
        }
        
        xmlXPathFreeObject(xpathObj);
      }
      xmlXPathFreeContext(xpathCtx);
    }
    xmlFreeDoc(doc);
  }
  return(res);
}

/**
 * copy_instruction:
 * @filename2:		the model XML filename, storing nodes used as model.
 * @xpathExpr2:	the xpath expression of the model node inside filename2 * 
 * @filename:		the input XML filename.
 * @xpathExpr:		the xpath expression of the parent node inside filename
 * 
 * Parses input XML files, evaluates XPaths expression and copy the node from
 * xpathExpr2 of filename2 as a child of xpathExpr in filename.
 * 
 * Returns 0 on success and a negative value otherwise.
 */
int copy_instruction(xmltoolAction type, const char* filename2, 
  const xmlChar* xpathExpr2, const char* filename, const xmlChar* xpathExpr, 
  int extented) {
    
  xmlDocPtr doc;
  xmlDocPtr doc2;
  xmlXPathContextPtr xpathCtx;
  xmlXPathContextPtr xpathCtx2; 
  xmlXPathObjectPtr xpathObj; 
  xmlXPathObjectPtr xpathObj2;
  int nodeNb=0;
  int nodeNb2=0;  
  
  xmlNodePtr cur=NULL;
  int res=-1;
  
  xmlIndentTreeOutput = 1;
  
  /* Load XML document */
  doc = xmlParseFile(filename);
  if (!doc) {
    error_msg( 0, filename);
  } else {
    doc2 = xmlParseFile(filename2);  
    if (!doc2) {
      error_msg( 0, filename2);
    } else {
    /* Create xpath evaluation context */
      xpathCtx = xmlXPathNewContext(doc);
      if (!xpathCtx) {
        error_msg( 1, filename);
      } else {
        xpathCtx2 = xmlXPathNewContext(doc2);
        if (!xpathCtx2) {
          error_msg( 1, filename2);
        } else {        
          /* Evaluate xpath expression */
          xpathObj = xmlXPathEvalExpression(xpathExpr, xpathCtx);
          if(!xpathObj){
            error_msg( 2, xpathExpr);
          } else {
            xpathObj2 = xmlXPathEvalExpression(xpathExpr2, xpathCtx2);
            if(!xpathObj2){
              error_msg( 2, xpathExpr2);
            } else {
              
              /* Print results */
              int nodeNb=(xpathObj->nodesetval) ? xpathObj->nodesetval->nodeNr : 0;
              int nodeNb2=(xpathObj2->nodesetval) ? xpathObj2->nodesetval->nodeNr : 0;
              
              if( (nodeNb==1) && (nodeNb2==1) ){
                cur=xmlCopyNode(xpathObj2->nodesetval->nodeTab[0],extented);
                switch(type) {
                  case COPY_AS_LASTCHILD:
                    xmlAddChild(xpathObj->nodesetval->nodeTab[0],cur);
                    break;
                  case COPY_AS_NEXTSIBLING:
                    xmlAddNextSibling(xpathObj->nodesetval->nodeTab[0],cur);
                    break;
                  case COPY_AS_PREVSIBLING:
                    xmlAddPrevSibling(xpathObj->nodesetval->nodeTab[0],cur);
                    break;
                }
                xmlSaveFormatFile(filename, doc, 1);
                res=0;
              }
              xmlXPathFreeObject(xpathObj2);
            }
            xmlXPathFreeObject(xpathObj);
          }
          xmlXPathFreeContext(xpathCtx2);
        } 
        xmlXPathFreeContext(xpathCtx); 
      } 
      xmlFreeDoc(doc2);
    }
    xmlFreeDoc(doc);
  }
  return(res);
}


int 
main(int argc, char **argv) {

  xmltoolAction type=NOTHING;
  int usageError;
  
  if (argc>3) {
    if (!strcmp(argv[1], "count") && (argc == 4))
      type=COUNT;
    else if (!strcmp(argv[1], "get") && (argc == 4))
      type=GET;
    else if (!strcmp(argv[1], "path") && (argc == 4))
      type=PATH;
    else if (!strcmp(argv[1], "remove") && (argc == 4))
      type=REMOVE;
    else if (!strcmp(argv[1], "set") && (argc == 5))
      type=SET;
    else if (!strcmp(argv[1], "copy_as_lastchild") && (argc == 7))
      type=COPY_AS_LASTCHILD;
    else if (!strcmp(argv[1], "copy_as_nextsibling") && (argc == 7))
      type=COPY_AS_NEXTSIBLING;
    else if (!strcmp(argv[1], "copy_as_prevsibling") && (argc == 7))
      type=COPY_AS_PREVSIBLING;
  }
  
  /* Init libxml */     
  xmlInitParser();
  LIBXML_TEST_VERSION
  
  /* Do the main job */
  if (type!=NOTHING){
    if((int)type>4) {
      usageError=copy_instruction( type, argv[2], BAD_CAST argv[3], argv[4], 
      BAD_CAST argv[5], (int)strtol(argv[6], (char **)NULL, 10));
    } else if((int)type<8) {
      usageError=execute_expression( type, argv[2], BAD_CAST argv[3], 
      (argc > 4) ? BAD_CAST argv[4] : NULL);
    }
  } else {
    usageError=-1;
  }
  
  if (usageError)
    usage(argv[0]);  
  
  xmlCleanupParser();
  return (usageError);
}
